import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import convert_config_to_qualified_name
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .utils import set_default


class ConfigSchema(ma.Schema):
    module = ma.fields.Str(metadata={"doc": "Module with app config"})
    extra_code = ma.fields.Str(
        metadata={"doc": "Extra code that will be pasted to app config"}
    )
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema),
        required=False,
        metadata={"doc": "Python imports"},
    )  # imports must be here as well as it is used on model's root (without field)

    class Meta:
        unknown = ma.RAISE


class ExtSchema(ma.Schema):
    alias = ma.fields.Str(
        metadata={
            "doc": "Alias under which the extension will be registered in setup.cfg and in app.extensions"
        }
    )
    module = ma.fields.Str(metadata={"doc": "Module with ext schema"})
    class_ = ma.fields.Str(
        attribute="class", data_key="class", metadata={"doc": "Extension class name"}
    )
    base_classes = ma.fields.Str(
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "A list of extension's base classes"},
    )
    skip = ma.fields.Boolean()
    extra_code = ma.fields.Str(
        metadata={"doc": "Extra code that will be pasted to app extension module"}
    )
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema),
        required=False,
        metadata={"doc": "Python imports"},
    )  # imports must be here as well as it is used on model's root (without field)

    class Meta:
        unknown = ma.RAISE


class AppModelComponent(DataTypeComponent):
    "flask application"
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        config = ma.fields.Nested(
            ConfigSchema, metadata={"doc": "Application config details"}
        )
        ext = ma.fields.Nested(
            ExtSchema, metadata={"doc": "Application extension details"}
        )

    def before_model_prepare(self, datatype, **kwargs):
        prefix = datatype.definition["module"]["prefix"]
        alias = datatype.definition["module"]["alias"]
        module = datatype.definition["module"]["qualified"]

        config = set_default(datatype, "config", {})

        config.setdefault("generate", True)
        config.setdefault("module", f"{module}.config")
        config.setdefault("extra_code", "")
        config.setdefault("imports", [])

        ext = set_default(datatype, "ext", {})

        ext.setdefault("generate", True)
        ext_module = ext.setdefault("module", f"{module}.ext")
        ext.setdefault("class", f"{ext_module}.{prefix}Ext")
        ext.setdefault("base-classes", [])
        ext.setdefault("extra_code", "")
        ext.setdefault("alias", alias)
        ext.setdefault("imports", [])
        convert_config_to_qualified_name(ext)



from marshmallow import fields

from .datatypes import DataType


class NumberDataType(DataType):
    facets = {
        "facet-class": "TermsFacet",
        "imports": [
            {"import": "invenio_records_resources.services.records.facets.TermsFacet"}
        ],
    }


class IntegerDataType(NumberDataType):
    model_type = "integer"

    ui = {
        "marshmallow": {
            "field-class": "ma_fields.Integer",
        }
    }
    marshmallow = {
        "field-class": "ma_fields.Integer",
    }
    json_schema = {"type": "integer"}

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Integer(required=False)
        exclusiveMinimum = fields.Integer(required=False)
        maximum = fields.Integer(required=False)
        exclusiveMaximum = fields.Integer(required=False)
        enum = fields.List(fields.Integer(), required=False)


class FloatDataType(NumberDataType):
    model_type = "float"

    ui = {
        "marshmallow": {
            "field-class": "ma_fields.Float",  # NOSONAR
        }
    }
    marshmallow = {
        "field-class": "ma_fields.Float",
    }
    json_schema = {"type": "number"}

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Float(required=False)
        exclusiveMinimum = fields.Float(required=False)
        maximum = fields.Float(required=False)
        exclusiveMaximum = fields.Float(required=False)
        enum = fields.List(fields.Float(), required=False)


class DoubleDataType(NumberDataType):
    model_type = "double"

    ui = {
        "marshmallow": {
            "field-class": "ma_fields.Float",
        }
    }
    marshmallow = {
        "field-class": "ma_fields.Float",
    }
    json_schema = {"type": "number"}

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Float(required=False)
        exclusiveMinimum = fields.Float(required=False)
        maximum = fields.Float(required=False)
        exclusiveMaximum = fields.Float(required=False)
        enum = fields.List(fields.Float(), required=False)


class BooleanDataType(DataType):
    model_type = "boolean"

    ui = {
        "marshmallow": {
            "field-class": "ma_fields.Boolean",
        }
    }
    marshmallow = {
        "field-class": "ma_fields.Boolean",
    }
    json_schema = {"type": "boolean"}
    facets = {
        "facet-class": "TermsFacet",
        "imports": [
            {"import": "invenio_records_resources.services.records.facets.TermsFacet"}
        ],
    }



import itertools

from libcst import Arg

from .common import IdentityMerger, MergerBase, PythonContext
from .mergers import expression_mergers


class CallMerger(MergerBase):
    def identity(self, context: PythonContext, node):
        return node.func

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        existing_args, existing_kwargs = self.extract_args(existing_node)
        new_args, new_kwargs = self.extract_args(new_node)
        args = []
        merger = ArgMerger()
        for e, n in itertools.zip_longest(existing_args, new_args):
            args.append(merger.merge(context, e, n))

        for k, e in existing_kwargs.items():
            n = new_kwargs.pop(k, None)
            args.append(merger.merge(context, e, n))

        for k, n in new_kwargs.items():
            args.append(merger.merge(context, None, n))

        args = [x for x in args if x is not context.REMOVED]

        return existing_node.with_changes(args=args)

    def extract_args(self, n):
        args = []
        kwargs = {}
        for arg in n.args:
            if isinstance(arg, Arg):
                if arg.keyword:
                    kwargs[arg.keyword.value] = arg
                else:
                    args.append(arg)
            else:
                raise RuntimeError(f"Unsupported clause in call args {type(arg)}")
        return args, kwargs


class ArgMerger(MergerBase):
    def identity(self, context: PythonContext, node):
        return node

    def merge_internal(self, context: PythonContext, existing_node, new_node):
        existing_value = existing_node.value
        new_value = new_node.value
        merger = expression_mergers.get(type(existing_value), IdentityMerger())
        merged_value = merger.merge(context, existing_value, new_value)
        return existing_node.with_changes(value=merged_value)



import dataclasses
from typing import Dict, List, Optional


@dataclasses.dataclass
class FacetDefinition:
    path: str
    dot_path: str
    searchable: bool
    imports: List[Dict[str, str]]
    field: Optional[str] = None

    def update(self, facet_section):
        self.imports.extend(facet_section.get("imports", []))
        if self.searchable is None:
            # searchable not set up, so take it from argument
            self.searchable = facet_section.get("searchable", None)

    def set_field(self, facet_section, arguments):
        field = facet_section.get("field")
        if field:
            self.field = field
        else:
            field_class = facet_section.get("facet-class")
            if field_class:
                self.field = f"{field_class}({', '.join(arguments)})"



from .json_base import JSONBaseBuilder


class ExtendBuilder(JSONBaseBuilder):
    TYPE = "extend"
    output_file_type = "json"

    def build_node(self, node):
        generated = self.generate(node)
        # on model, move the marshmallow class into base_classes
        self._process_top_level_marshmallow(generated)
        generated["type"] = "object"
        self.output.merge(generated)

    def generate(self, node):
        ret = {**node.definition}
        ret.pop("properties", None)
        ret.pop("items", None)

        self._process_marshmallow_def(ret.setdefault("marshmallow", {}))
        self._process_marshmallow_def(
            ret.setdefault("ui", {}).setdefault("marshmallow", {})
        )

        if getattr(node, "children", None):
            properties = ret.setdefault("properties", {})
            for k, v in node.children.items():
                v = self.generate(v)
                properties[k] = v
        if getattr(node, "item", None):
            ret["items"] = self.generate(node.item)
        return ret

    def _process_marshmallow_def(self, marshmallow):
        marshmallow.update({"read": False, "write": False})
        if "class" in marshmallow:
            # already generated - if user wants to override this, he has to set schema-class and generate
            marshmallow["generate"] = False

    def _process_top_level_marshmallow(self, model):
        marshmallow = model.pop("marshmallow")
        ui = model.pop("ui")

        for k in list(model.keys()):
            if k.endswith("-class"):
                prefix = k[:-6]
                model[f"{prefix}-bases"] = [model.pop(k)]
            elif k not in ("type", "properties"):
                model.pop(k)  # pop all other stuff

        model["marshmallow"] = {"base-classes": [marshmallow["class"]]}
        model["ui"] = {"marshmallow": {"base-classes": [ui["marshmallow"]["class"]]}}

    def begin(self, current_model, schema):
        super(JSONBaseBuilder, self).begin(current_model, schema)
        self.output = self.builder.get_output(self.output_file_type, "model.json5")

    def finish(self):
        # force clean output
        super().finish()
        self.output.force_clean_output()



from pathlib import Path
from typing import Any, Dict, Iterable, List, Type, Union

from .builders import OutputBuilder
from .fs import AbstractFileSystem, FileSystem
from .outputs import OutputBase
from .schema import ModelSchema
from .utils.import_class import import_class


class ModelBuilder:
    """
    Processes a model file and generates/updates sources for the model
    """

    output_classes: List[Type[OutputBase]]
    """
    Mapping between output type and its handler class
    """

    filtered_output_classes: Dict[str, Type[OutputBase]]
    """
    Filtered output classes by settings.plugins.disabled|enabled
    """

    output_builder_classes: List[Type[OutputBuilder]]
    """
    A list of extension classes to be used in build. 
    """

    output_builders: List[OutputBuilder]
    """
    A list of output_builders. Each extension is responsible for generating one or more files
    """

    outputs: Dict[Path, OutputBase]
    """
    Mapping between concrete output (file path relative to output dir) and instance of builder class
    """

    filesystem: AbstractFileSystem

    overwrite: bool
    """
    If true, overwrite already existing files. If false, perform merge
    """

    def __init__(
        self,
        outputs: List[Type[OutputBase]] = (),
        output_builders: List[Type[OutputBuilder]] = (),
        filesystem=FileSystem(),
        overwrite=False,
    ):
        """
        Initializes the builder

        :param output_builders:          A list of extension classes to use in builds
        :param outputs:     List of file builder classes that generate files
        """
        self.output_builder_classes = [*output_builders]
        for o in outputs:
            assert o.TYPE, f"output_type not set up on class {o}"
        self.output_classes = [*(outputs or [])]
        self.outputs = {}
        self.filtered_output_classes = {o.TYPE: o for o in self.output_classes}
        self.filesystem = filesystem
        self.overwrite = overwrite

    def get_output(self, output_type: str, path: Union[str, Path]):
        """
        Given a path, instantiate file builder on the path with the given output type
        and return it. If the builder on the path has already been requested, return
        the same instance of the builder.

        :param output_type: @see FileBuilder.output_type
        :param path: relative path to output_dir, set in build()
        :return:    instance of FileBuilder for the path
        """
        if not isinstance(path, Path):
            path = Path(path)
        path = self.output_dir.joinpath(path)

        output = self.outputs.get(path, None)
        if output:
            assert output_type == self.outputs[path].TYPE
        else:
            output = self.filtered_output_classes[output_type](self, path)
            output.begin()
            self.outputs[path] = output
        return output

    # main entry point
    def build(
        self,
        model: ModelSchema,
        profile: str,
        model_path: List[str],
        output_dir: Union[str, Path],
        context: Dict[str, Any] = None,
    ):
        """
        compile the schema to output directory

        :param model:       the model schema
        :param profile:     the profile under which the builder runs
        :param model_path:  path within the schema that will be converted to datatype and used by output builders
        :param output_dir:  output directory where to put generated files
        :param context:     extra context supplied to datatype preparation
        :return:            the outputs (self.outputs)
        """

        # deep copy the model
        if self.overwrite:
            self.filesystem.overwrite = True

        self.set_schema(model)
        self.filtered_output_classes = {
            o.TYPE: o
            for o in self._filter_classes(self.output_classes, model.schema, "output")
        }
        self.output_dir = Path(output_dir).absolute()
        self.outputs = {}

        self._run_output_builders(model, profile, model_path, context or {})

        self._save_outputs()

        return self.outputs

    def _run_output_builders(
        self, model: ModelSchema, profile: str, model_path: Iterable[str], context
    ):
        output_builder_class: Type[OutputBuilder]
        for output_builder_class in self._filter_classes(
            self.output_builder_classes, model.schema, "builder"
        ):
            output_builder = output_builder_class(builder=self)
            current_model = model.get_schema_section(profile, model_path, context)
            output_builder.build(current_model=current_model, schema=model.schema)

    def _save_outputs(self):
        for output in sorted(self.outputs.values(), key=lambda x: x.path):
            output.finish()
            if output.executable:
                self.filesystem.make_executable(output.path)

    def set_schema(self, schema):
        self.schema = schema
        self.settings = schema.settings

    # private methods

    def _filter_classes(self, classes: List[Type[object]], model, plugin_type):
        if "plugins" not in model or plugin_type not in model["plugins"]:
            return classes
        plugin_config = model["plugins"][plugin_type]

        disabled = plugin_config.get("disable", [])
        enabled = plugin_config.get("enable", [])
        included = plugin_config.get("include", [])

        if included:
            enabled = [*enabled]  # will be adding inclusions so make a copy
            classes = [*classes]
            for incl in included:
                class_type = import_class(incl)
                classes.append(class_type)
                if enabled and class_type.TYPE not in enabled:
                    enabled.append(class_type.TYPE)

        if disabled == "__all__":
            ret = []
        else:
            ret = [c for c in classes if c.TYPE not in disabled]

        if enabled:
            ret.extend([c for c in classes if c.TYPE in enabled])
        return ret



from typing import List

from oarepo_model_builder.datatypes import DataType

from ..datatypes.components.facets import FacetDefinition
from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordSearchFacetsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_facets"
    section = "facets"
    template = "record-facets"

    def build_node(self, node: DataType):
        # everything is done in finish
        pass

    def finish(self, **extra_kwargs):
        facets: List[FacetDefinition] = []
        for node in self.current_model.deep_iter():
            facets.extend(node.section_facets.config["facets"])
        package = self.current_model.definition["facets"]["module"]

        imports = []
        for f in facets:
            imports.extend(f.imports)

        return super().finish(
            current_package_name=package,
            facets=facets,
            imports=imports,
            **extra_kwargs,
        )



import os

import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import module_to_path, parent_module
from oarepo_model_builder.validation.utils import PermissiveSchema

from .defaults import DefaultsModelComponent
from .record import RecordModelComponent
from .utils import set_default


class JSONSchema(ma.Schema):
    generate = ma.fields.Bool(
        metadata={"doc": "Generate json schema (default is true)"}
    )
    alias = ma.fields.Str(
        metadata={"doc": "Alias under which jsonschema is referenced in setup.cfg"}
    )
    version = ma.fields.Str(metadata={"doc": "Schema version"})
    name = ma.fields.Str(metadata={"doc": "Schema name"})
    module = ma.fields.Str(metadata={"doc": "Schema module"})
    file_ = ma.fields.Str(
        data_key="file", attribute="file", metadata={"doc": "Path to schema file"}
    )

    # json schema
    template = ma.fields.Nested(
        PermissiveSchema,
        metadata={"doc": "Template that will be merged with the schema"},
    )
    skip = ma.fields.Boolean()

    class Meta:
        unknown = ma.RAISE


class JSONSchemaModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, RecordModelComponent]

    class ModelSchema(ma.Schema):
        jsonschema = ma.fields.Nested(
            JSONSchema,
            data_key="json-schema-settings",
            attribute="json-schema-settings",
            metadata={
                "doc": "JSON Schema section of the model. Properties will be generated automatically"
            },
        )

    def before_model_prepare(self, datatype, **kwargs):
        prefix_snake = datatype.definition["module"]["prefix-snake"]
        alias = datatype.definition["module"]["alias"]
        records_path = module_to_path(
            parent_module(datatype.definition["record"]["module"])
        )

        json_schema = set_default(datatype, "json-schema-settings", {})
        json_schema.setdefault("generate", True)
        json_schema.setdefault("alias", alias)
        version = json_schema.setdefault(
            "version", datatype.schema.settings.get("version", "1.0.0")
        )
        json_schema.setdefault(
            "module",
            f'{parent_module(datatype.definition["record"]["module"])}.jsonschemas',
        )
        schema_name = json_schema.setdefault(
            "name",
            f"{prefix_snake}-{version}.json",
        )

        json_schema.setdefault(
            "file",
            os.path.join(records_path, "jsonschemas", schema_name),
        )



from typing import List

from ... import DataType, ObjectDataType
from .graph import MarshmallowField, MarshmallowReference
from .object import ObjectMarshmallowMixin
from .ui_field import UIMarshmallowComponent


class UIObjectMarshmallowComponent(ObjectMarshmallowMixin, UIMarshmallowComponent):
    eligible_datatypes = [ObjectDataType]

    def ui_marshmallow_register_class_names(
        self, *, datatype, classes, marshmallow_module, **__kwargs
    ):
        self._register_class_name(
            datatype,
            datatype.section_ui.config.setdefault("marshmallow", {}),
            classes,
            marshmallow_module,
        )

    def ui_marshmallow_build_class_name_existing(
        self, *, datatype, classes, marshmallow_module, **__kwargs
    ):
        if datatype.section_ui.config.get("marshmallow", {}).get("class"):
            self._build_class_name(
                datatype,
                datatype.section_ui.config.setdefault("marshmallow", {}),
                datatype.definition.setdefault("ui", {}).setdefault("marshmallow", {}),
                classes,
                marshmallow_module,
                datatype.section_ui.fingerprint,
                "UISchema",
            )

    def ui_marshmallow_build_class_name_new(
        self, *, datatype, classes, marshmallow_module, **__kwargs
    ):
        if not datatype.section_ui.config.get("marshmallow", {}).get("class"):
            self._build_class_name(
                datatype,
                datatype.section_ui.config.setdefault("marshmallow", {}),
                datatype.definition.setdefault("ui", {}).setdefault("marshmallow", {}),
                classes,
                marshmallow_module,
                datatype.section_ui.fingerprint,
                "UISchema",
            )

    def ui_marshmallow_build_class(self, *, datatype, classes, **__kwargs):
        self._build_class(
            datatype,
            datatype.section_ui.config.setdefault("marshmallow", {}),
            datatype.section_ui.children,
            "ui_marshmallow_field",
            classes,
        )

    def ui_marshmallow_field(
        self, datatype: DataType, *, fields: List[MarshmallowField], **__kwargs
    ):
        section = datatype.section_ui
        f = []
        super().ui_marshmallow_field(datatype, fields=f)
        fld: MarshmallowField = f[0]
        fld.reference = MarshmallowReference(
            reference=section.config["marshmallow"]["class"]
        )
        fields.append(fld)

    def _marshmallow_field_arguments(self, datatype, section, marshmallow, field_name):
        return [
            "__reference__",
            *super()._marshmallow_field_arguments(
                datatype, section, marshmallow, field_name
            ),
        ]



import importlib.resources
import sys
from functools import reduce
from importlib import import_module
from pathlib import Path

import importlib_metadata

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.schema import ModelSchema, remove_star_keys


def create_builder_from_entrypoints(profile="record", **kwargs):
    # output classes do not depend on profile
    output_classes = load_entry_points_list("oarepo_model_builder.outputs", None)
    builder_classes = load_entry_points_list("oarepo_model_builder.builders", profile)

    return ModelBuilder(
        output_builders=builder_classes,
        outputs=output_classes,
        **kwargs,
    )


def load_entry_points_dict(name):
    return {
        ep.name: ep.load()
        for ep in importlib_metadata.entry_points().select(group=name)
    }


def load_entry_points_list(name, profile):
    ret = []
    loaded = {}
    group_name = f"{name}.{profile}" if profile else name
    load_from_entry_point_internal(name, group_name, loaded, ret)
    # inherit
    inherit_group = f"{name}.{profile}.inherit" if profile else f"{name}.inherit"
    for ep in importlib_metadata.entry_points().select(group=inherit_group):
        inherit_from = ep.value
        load_from_entry_point_internal(name, inherit_from, loaded, ret)
    ret.sort()
    return [x[1] for x in ret]


def load_from_entry_point_internal(name, group_name, loaded_keys, loaded_entries):
    for ep in importlib_metadata.entry_points().select(group=group_name):
        if ep.name in loaded_keys:
            print(
                f"WARNING: Entry point {ep.name} has already been registered to group {name}. "
                f"Previous value {loaded_keys[ep.name]}, new ignored value {ep.value}"
            )
            continue
        loaded_entry_point = ep.load()
        loaded_entries.append((ep.name, loaded_entry_point))
        loaded_keys[ep.name] = ep.value


def load_model_from_entrypoint(ep: importlib_metadata.EntryPoint):
    def load(schema):
        try:
            loaded_schema = ep.load()
        except:  # NOSONAR intentionally broad
            module = import_module(ep.module)
            split_attr = ep.attr.split(".")
            fn = f"{split_attr[-2]}.{split_attr[-1]}"
            if len(split_attr) > 2:
                fn = reduce(lambda x, y: Path(x) / Path(y), split_attr[:-2]) / fn
            module_path = getattr(module, "__path__", [])
            if module_path:
                full_fn = Path(module_path[0]) / fn
            else:
                full_fn = fn
            content = importlib.resources.open_text(module, fn, encoding="utf-8").read()
            loaded_schema = schema._load(full_fn, content=content)

        remove_star_keys(loaded_schema)
        return loaded_schema

    return load


def load_included_models_from_entry_points():
    ret = {}
    for ep in importlib_metadata.entry_points().select(group="oarepo.models"):
        ret[ep.name] = load_model_from_entrypoint(ep)
    return ret


def load_model(
    model_filename,
    configs=(),
    black=True,
    isort=True,
    autoflake=True,
    sets=(),
    model_content=None,
    extra_included=None,
    merged_models=None,
):
    loaders = load_entry_points_dict("oarepo_model_builder.loaders")
    included_models = load_included_models_from_entry_points()
    if extra_included:
        included_models.update(extra_included)
    schema = ModelSchema(
        model_filename,
        content=model_content,
        loaders=loaders,
        included_models=included_models,
        merged_models=merged_models,
    )
    for config in configs:
        load_config(schema, config, loaders)
    for s in sets:
        k, v = s.split("=", 1)
        schema.schema[k] = v
    check_plugin_packages(schema)
    if "python" not in schema.settings:
        schema.settings["python"] = {}
    schema.settings["python"]["use-isort"] = isort
    schema.settings["python"]["use-black"] = black
    schema.settings["python"]["use-autoflake"] = autoflake
    return schema


def load_config(schema, config, loaders):
    old_loaders = schema.loaders
    schema.loaders = loaders
    try:
        loaded_file = schema._load(config)
        schema.merge(loaded_file)
    finally:
        schema.loaders = old_loaders


def check_plugin_packages(schema):
    try:
        required_packages = schema.schema.plugins.packages
    except AttributeError:
        return
    import subprocess

    import pkg_resources

    known_packages = set(d.project_name for d in pkg_resources.working_set)
    unknown_packages = [rp for rp in required_packages if rp not in known_packages]
    if unknown_packages:
        if (
            input(
                f'Required packages {", ".join(unknown_packages)} are missing. '
                f"Should I install them for you via pip install? (y/n) "
            )
            == "y"
        ):
            if subprocess.call(["pip", "install", *unknown_packages]):
                sys.exit(1)
            print("Installed required packages, please run this command again")
        sys.exit(1)



from typing import List

from ..datatypes.components.facets import FacetDefinition
from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordSearchOptionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_search_options"
    section = "search-options"
    template = "record-search-options"

    def finish(self, **extra_kwargs):
        facets: List[FacetDefinition] = []
        for node in self.current_model.deep_iter():
            facets.extend(node.section_facets.config["facets"])
        search_data = []
        for f in facets:
            search_data.append({f.path: "facets." + f.path})
        extra_kwargs["search_data"] = search_data
        super().finish(**extra_kwargs)



from pathlib import Path

from ..builders import OutputBuilder
from ..builders.utils import ensure_parent_modules
from ..outputs.cfg import CFGOutput
from ..utils.python_name import module_to_path, split_package_base_name


class InvenioRecordMetadataAlembicSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_record_metadata_alembic_setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")
        alembic_module_parent, alembic_module = split_package_base_name(
            self.current_model.definition["record-metadata"]["alembic"]
        )
        output.add_entry_point(
            "invenio_db.alembic",
            self.current_model.definition["record-metadata"]["alias"],
            f"{alembic_module_parent}:{alembic_module}",
        )

        python_path = (
            Path(
                module_to_path(
                    self.current_model.definition["record-metadata"]["alembic"]
                )
            )
            / "__init__.py"
        )
        # create parent modules if they do not exist
        ensure_parent_modules(
            self.builder, python_path, max_depth=len(python_path.parts)
        )

        # and create empty __init__.py
        self.builder.get_output("python", python_path)



from marshmallow import fields

from ..datatypes import DataType
from .object import FieldSchema


class ArrayDataType(DataType):
    model_type = "array"

    ui = {
        "marshmallow": {
            "field-class": "ma_fields.List",
        }
    }
    marshmallow = {
        "field-class": "ma_fields.List",
    }
    json_schema = {"type": "array"}

    class ModelSchema(DataType.ModelSchema):
        items = fields.Nested(FieldSchema)
        uniqueItems = fields.Boolean(required=False)
        minItems = fields.Integer(required=False)
        maxItems = fields.Integer(required=False)

    def prepare(self, context):
        from ..datatypes import datatypes

        items = self.definition.get("items", {})
        self.item = datatypes.get_datatype(self, items, None, self.model, self.schema)
        self.item.prepare(context)
        super().prepare(context)

    def deep_iter(self):
        yield from super().deep_iter()
        yield from self.item.deep_iter()



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioAPIViewsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_api_views"
    section = "api-blueprint"
    template = "api-views"



import json
from typing import Callable

import faker
import faker.providers
from faker import Faker

from oarepo_model_builder.datatypes import (
    ArrayDataType,
    DataType,
    ObjectDataType,
    Section,
)

from ..builder import ModelBuilder
from ..builders.json_base import JSONBaseBuilder
from ..entrypoints import load_entry_points_list


class SampleDataGenerator(faker.Generator):
    def __init__(self, **config):
        super().__init__(**config)
        self.formatters = set()

    def set_formatter(self, name: str, method: Callable) -> None:
        self.formatters.add(name)
        return super().set_formatter(name, method)


# provider for Faker
class Provider:
    def __init__(self, generator) -> None:
        self.generator = generator

    def random_float(self):
        rnd = self.generator.random.randrange(-100, 100 + 1, 1)
        return rnd / 10

    def random_boolean(self):
        rnd = self.generator.random.randrange(0, 2, 1)
        return rnd == 1

    def sample_object(self):
        return {
            self.generator.word(): self.generator.word()
            for _ in range(self.generator.random.randrange(1, 5, 1))
        }

    def constant(self, value=None):
        return value

    def language_dict(self):
        return {lang: Faker(locale=lang).sentence() for lang in ("cs", "en")}


SKIP = "skip"


class SampleDataBuilder(JSONBaseBuilder):
    TYPE = "script_sample_data"
    output_file_type = "yaml"
    output_file_name = ["sample", "file"]
    parent_module_root_name = "jsonschemas"

    def __init__(self, builder: ModelBuilder):
        super().__init__(builder)
        self.generator = SampleDataGenerator()
        from faker.config import PROVIDERS

        self.faker = faker.Faker(
            generator=self.generator,
            providers=[
                "oarepo_model_builder.invenio.invenio_script_sample_data",
                *PROVIDERS,
            ],
        )

        self.sample_data_providers = load_entry_points_list(
            "oarepo_model_builder.sample_data_providers", profile=None
        ) + [faker_provider]

    def build_node(self, node: DataType):
        if not self.output.created:
            return

        sample: Section = node.section_sample
        for __ in range(sample.config.get("count", 10)):
            self.output.next_document()
            generated = self.generate_sample_for_node_and_children(node)
            self.output.merge(generated)

    def generate_sample_for_node_and_children(self, node):
        sample_section, sample = get_oarepo_sample(node)
        if "sample" in sample:
            return sample["sample"]
        if sample.get("skip"):
            return SKIP

        if isinstance(node, ObjectDataType):
            ret = {}
            for k, v in sample_section.children.items():
                v = self.generate_sample_for_node_and_children(v)
                if v is not SKIP:
                    ret[k] = v
        elif isinstance(node, ArrayDataType):
            count = sample.get("count")
            if count is None:
                count = self.faker.random_int(1, 5)
            ret = {}
            for __ in range(count):
                v = self.generate_sample_for_node_and_children(sample_section.item)
                if v is not SKIP:
                    ret[json.dumps(v, sort_keys=True)] = v
            ret = list(ret.values())
        else:
            ret = self.generate_fake(node, sample)
        return ret

    def generate_fake(self, node: DataType, config):
        params = {}
        method = None
        params = config.get("params", params)

        for provider in self.sample_data_providers:
            ret = provider(self.faker, node.schema.schema["settings"], node, params)
            if ret is not SKIP:
                return ret
        return SKIP


def get_oarepo_sample(node):
    sample_section = node.section_sample
    sample = sample_section.config

    if isinstance(sample, dict):
        return sample_section, sample
    if isinstance(sample, str):
        return sample_section, {"faker": "constant", "params": {"value": sample}}
    if isinstance(sample, (list, dict)):
        if not node.key:
            # array
            return sample_section, {
                "faker": "random_elements",
                "params": {"elements": sample, "unique": True},
            }
        else:
            # element
            return sample_section, {
                "faker": "random_element",
                "params": {"elements": sample},
            }

    return sample_section, {}


def faker_provider(faker, settings, node, params):
    __, config = get_oarepo_sample(node)
    method = config.get("faker")
    if not method:
        if node.key in faker.formatters:
            method = node.key
        else:
            data_type = node.model_type
            if data_type == "integer":
                method = "random_int"
            elif data_type == "number":
                method = "random_float"
            elif data_type == "double":
                method = "random_float"
            elif data_type == "float":
                method = "random_float"
            elif data_type == "boolean":
                method = "random_boolean"
            elif data_type == "date":
                method = "date"
            elif data_type == "object":
                # it is an object with unknown properties, return a sample object
                method = "sample_object"
            elif data_type in ("string", "keyword", "fulltext", "fulltext+keyword"):
                method = "sentence"
            else:
                print(
                    f"Warning: do not know how to generate sample data for {data_type} at path {node.path}, using plain string rule"
                )
                method = "sentence"
    return getattr(faker, method)(**params)



from .array import ArrayDataType
from .flatten import FlattenDataType
from .nested import NestedDataType
from .object import ObjectDataType

__all__ = ["ObjectDataType", "NestedDataType", "FlattenDataType", "ArrayDataType"]



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioAPPViewsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_app_views"
    section = "app-blueprint"
    template = "app-views"



from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioRecordSearchSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_record_search_setup_cfg"

    def finish(self):
        super().finish()
        if self.current_model.definition["mapping-settings"].get("skip"):
            return

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        output.add_entry_point(
            "invenio_search.mappings",
            self.current_model.definition["mapping-settings"]["alias"],
            self.current_model.definition["mapping-settings"]["module"],
        )



from ..datatypes import DataType


class FlattenDataType(DataType):
    model_type = "flattened"

    ui = {
        "marshmallow": {
            "field-class": "ma_fields.Raw",
        }
    }
    marshmallow = {
        "field-class": "ma_fields.Raw",
    }
    json_schema = {"type": "object"}
    mapping = {"enabled": False, "type": "object"}



from .datatypes import DataType


class BaseDateDataType(DataType):
    marshmallow = {"field-class": "ma_fields.String"}

    facets = {
        "facet-class": "DateTimeFacet",
        "imports": [{"import": "oarepo_runtime.facets.date.DateTimeFacet"}],
    }


class DateDataType(BaseDateDataType):
    model_type = "date"
    ui = {
        "marshmallow": {
            "field-class": "l10n.LocalizedDate",
            "imports": [
                {"import": "oarepo_runtime.ui.marshmallow", "alias": "l10n"}  # NOSONAR
            ],
        }
    }
    marshmallow = {
        "field-class": "ma_fields.String",
        "validators": ["validate_date('%Y-%m-%d')"],
        "imports": [{"import": "oarepo_runtime.validation.validate_date"}],
    }
    mapping = {"type": "date", "format": "basic_date||strict_date"}
    json_schema = {"type": "string", "format": "date"}


class TimeDataType(BaseDateDataType):
    model_type = "time"

    ui = {
        "marshmallow": {
            "field-class": "l10n.LocalizedTime",
            "imports": [{"import": "oarepo_runtime.ui.marshmallow", "alias": "l10n"}],
        }
    }
    marshmallow = {
        "field-class": "ma_fields.String",
        "validators": ["validate_date('%H:%M:%S')"],
        "imports": [{"import": "oarepo_runtime.validation.validate_date"}],
    }
    mapping = {
        "type": "date",
        "format": "strict_time||strict_time_no_millis||basic_time||basic_time_no_millis",
    }
    json_schema = {"type": "string", "format": "time"}


class DateTimeDataType(BaseDateDataType):
    model_type = "datetime"

    ui = {
        "marshmallow": {
            "field-class": "l10n.LocalizedDateTime",
            "imports": [{"import": "oarepo_runtime.ui.marshmallow", "alias": "l10n"}],
        }
    }
    marshmallow = {
        "field-class": "ma_fields.String",
        "validators": ["validate_datetime"],
        "imports": [{"import": "oarepo_runtime.validation.validate_datetime"}],
    }
    mapping = {
        "type": "date",
        "format": "strict_date_time||strict_date_time_no_millis||basic_date_time||basic_date_time_no_millis||basic_date||strict_date",
    }
    json_schema = {"type": "string", "format": "date-time"}


class EDTFDataType(BaseDateDataType):
    model_type = "edtf"

    ui = {
        "marshmallow": {
            "field-class": "l10n.LocalizedEDTF",
            "imports": [{"import": "oarepo_runtime.ui.marshmallow", "alias": "l10n"}],
        }
    }
    marshmallow = {
        "field-class": "ma_fields.String",
        "validators": ["CachedMultilayerEDTFValidator(types=(EDTFDate,))"],
        "imports": [
            {"import": "oarepo_runtime.validation.CachedMultilayerEDTFValidator"},
            {"import": "edtf.Date", "alias": "EDTFDate"},
        ],
    }
    mapping = {
        "type": "date",
        "format": "strict_date_time||strict_date_time_no_millis||strict_date||yyyy-MM||yyyy",
    }
    json_schema = {"type": "string", "format": "date-time"}


class EDTFIntervalType(BaseDateDataType):
    model_type = "edtf-interval"

    ui = {
        "marshmallow": {
            "field-class": "l10n.LocalizedEDTFInterval",
            "imports": [{"import": "oarepo_runtime.ui.marshmallow", "alias": "l10n"}],
        }
    }
    marshmallow = {
        "field-class": "ma_fields.String",
        "validators": ["CachedMultilayerEDTFValidator(types=(EDTFInterval,))"],
        "imports": [
            {"import": "oarepo_runtime.validation.CachedMultilayerEDTFValidator"},
            {"import": "edtf.Interval", "alias": "EDTFInterval"},
        ],
    }
    mapping = {
        "type": "date_range",
        "format": "strict_date_time||strict_date_time_no_millis||strict_date||yyyy-MM||yyyy",
    }
    json_schema = {"type": "string", "format": "date-time"}



from pathlib import Path
from typing import Any, Dict, List, Union

import json5

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.entrypoints import create_builder_from_entrypoints, load_model
from oarepo_model_builder.fs import InMemoryFileSystem
from oarepo_model_builder.profiles import Profile
from oarepo_model_builder.profiles.extend import ExtendProfile
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.utils.deepmerge import deepmerge
from oarepo_model_builder.utils.dict import dict_get


class RecordProfile(Profile):
    default_model_path = ("record",)

    def build(
        self,
        model: ModelSchema,
        profile: str,
        model_path: List[str],
        output_directory: Union[str, Path],
        builder: ModelBuilder,
        **kwargs,
    ):
        current_model = dict_get(model.schema, model_path)
        if "extend" in current_model:
            self.handle_extend(
                current_model["extend"],
                model,
                profile,
                model_path,
                current_model,
                builder,
            )
        return super().build(
            model, profile, model_path, output_directory, builder, **kwargs
        )

    def handle_extend(
        self,
        extended_schema: str,
        model: ModelSchema,
        profile: str,
        model_path: List[str],
        current_model: Dict[str, Any],
        builder: ModelBuilder,
    ):
        """
        extending the model means:
            1. extended schema is read as a schema. If there is a json path behind it,
               it is used to select the model. by default /model is taken
            2. The model will be processed through "extend" profile with
               the output file "model.json5" stored in memory:
                  * properties ending with "-class" on model will be turned to array and suffix changed to -base-classes
                  * jsonschema and mapping will not be touched
                  * marshmallow will have all the properties marked as {read: false, write: false} so that they are not generated
            3. the resulting model will be merged as if use: was used
        """

        loaded_schema = {"model": {"use": [extended_schema]}}

        extended_model = load_model(
            extended_schema.split("#", maxsplit=1)[0],
            configs=[],
            black=False,
            isort=False,
            autoflake=False,
            sets=[],
            extra_included=model.included_schemas,
            model_content=loaded_schema,
        )
        extended_model.model_field = "model"

        fs = InMemoryFileSystem()
        builder = create_builder_from_entrypoints(
            profile="extend",
            overwrite=builder.overwrite,
            filesystem=fs,
        )

        ExtendProfile().build(extended_model, profile, model_path, "", builder)

        loaded_model = json5.loads(fs.read("model.json5"))
        deepmerge(current_model, loaded_model)



from typing import List

import marshmallow as ma
from marshmallow import fields

from oarepo_model_builder.datatypes import DataType, ObjectDataType, datatypes
from oarepo_model_builder.utils.absolute_class import convert_to_absolute_class_name
from oarepo_model_builder.utils.python_name import (
    base_name,
    convert_name_to_python_class,
    package_name,
    qualified_name,
)
from oarepo_model_builder.validation import InvalidModelException
from oarepo_model_builder.validation.utils import ImportSchema

from .field import (
    Import,
    MarshmallowField,
    PropertyMarshmallowSchema,
    RegularMarshmallowComponent,
)
from .graph import MarshmallowClass, MarshmallowReference


class ExtraField(ma.Schema):
    name = fields.String(required=True, metadata={"doc": "Name (lhs) of the field"})
    value = fields.String(
        required=True, metadata={"doc": "Literal definition (rhs) of the field"}
    )


class ObjectMarshmallowExtraSchema(ma.Schema):
    imports = fields.List(
        fields.Nested(ImportSchema),
        required=False,
        metadata={"doc": "Python imports that will be put to marshmallow file"},
    )  # imports must be here as well as it is used on model's root (without field)
    module = ma.fields.String(metadata={"doc": "Class module"})
    generate = fields.Boolean(
        required=False,
        metadata={"doc": "Generate the marshmallow class (default is true)"},
    )
    schema_class = fields.String(
        data_key="class",
        attribute="class",
        required=False,
        allow_none=True,
        metadata={"doc": "The name of the marshmallow class"},
    )
    base_classes = fields.List(
        fields.String(),
        data_key="base-classes",
        attribute="base-classes",
        required=False,
        metadata={"doc": "List of marshmallow base classes"},
    )
    extra_fields = fields.List(
        fields.Nested(ExtraField),
        required=False,
        data_key="extra-fields",
        attribute="extra-fields",
        metadata={"doc": "Extra fields to generate into the marshmallow class"},
    )
    skip = fields.Boolean()


class ObjectMarshmallowSchema(PropertyMarshmallowSchema, ObjectMarshmallowExtraSchema):
    pass


class ObjectMarshmallowMixin:
    def _register_class_name(
        self, datatype, marshmallow_config, classes, marshmallow_module
    ):
        schema_class = marshmallow_config.get("class")
        if not schema_class:
            return
        schema_class = convert_to_absolute_class_name(schema_class, marshmallow_module)
        classes[schema_class].append(
            (datatype, marshmallow_config.get("generate", True))
        )

    def _build_class_name(
        self,
        datatype,
        marshmallow_config,
        definition_marshmallow,
        classes,
        marshmallow_module,
        fingerprint,
        suffix,
    ):
        schema_class = marshmallow_config.get("class")
        generate = marshmallow_config.get("generate", True)

        if schema_class:
            qualified_schema_class = qualified_name(marshmallow_module, schema_class)
            if qualified_schema_class != schema_class:
                marshmallow_config["class"] = qualified_schema_class
                definition_marshmallow["class"] = qualified_schema_class
                schema_class = qualified_schema_class
            if not generate:
                if fingerprint not in classes:
                    classes[fingerprint] = schema_class
                return

        fingerprint_schema_class = classes.get(fingerprint)
        if fingerprint_schema_class:
            # reuse, do not generate again, even if schema_class was specified
            schema_class = fingerprint_schema_class
            marshmallow_config["generate"] = False
            definition_marshmallow["generate"] = False
        else:
            schema_class = self._find_unique_schema_class(
                schema_class, datatype, classes, marshmallow_module, suffix
            )
            marshmallow_config["generate"] = True
            definition_marshmallow["generate"] = True
            classes[fingerprint] = schema_class

        marshmallow_config["class"] = schema_class
        definition_marshmallow["class"] = schema_class

        classes[schema_class].append((datatype, marshmallow_config["generate"]))

    def _find_unique_schema_class(
        self, original_schema_class, datatype, classes, marshmallow_module, suffix
    ):
        schema_class_list = []
        if original_schema_class:
            # insert dummy datatype with class name
            datatype = DataType(
                parent=datatype,
                definition={},
                key=base_name(original_schema_class),
                model=datatype.model,
                schema=datatype.schema,
            )
            marshmallow_module = (
                package_name(original_schema_class) or marshmallow_module
            )
            if original_schema_class.endswith(suffix):
                suffix = ""

        while datatype:
            if not datatype.key:
                datatype = datatype.parent
                schema_class_list.insert(0, "Item")
                continue

            schema_class_list.insert(0, datatype.key)

            schema_class = (
                convert_name_to_python_class("-".join(x for x in schema_class_list))
                + suffix
            )
            schema_class = convert_to_absolute_class_name(
                schema_class, marshmallow_module
            )
            if schema_class not in classes:
                return schema_class

            datatype = datatype.parent

        # generate unique class name (if duplicates are found) by appending a number
        class_name = schema_class[: -len(suffix)]
        for i in range(1, 100):
            schema_class = f"{class_name}{i}{suffix}"
            if schema_class not in classes:
                return schema_class

        raise InvalidModelException(
            f"Too many marshmallow classes with name {class_name}{suffix}, defined on path {datatype.path}. "
            "Please specify your own class name for marshmallow at this path"
        )

    def _build_class(
        self, datatype, marshmallow, children, field_generator, classes  # NOSONAR
    ):
        fields = []
        for _, c in sorted(children.items()):
            datatypes.call_components(c, field_generator, fields=fields)
        extra_fields = [
            MarshmallowField(f["name"], f["value"])
            for f in marshmallow.get("extra-fields", [])
        ]
        fields = [*fields, *extra_fields]
        fields.sort(key=lambda x: (not x.key.startswith("_"), x.key))
        classes.append(
            MarshmallowClass(
                class_name=marshmallow["class"],
                base_classes=marshmallow.get("base-classes", []) or ["ma.Schema"],
                imports=Import.from_config(marshmallow.get("imports", [])),
                fields=fields,
                strict=True,
            )
        )


class ObjectMarshmallowComponent(ObjectMarshmallowMixin, RegularMarshmallowComponent):
    eligible_datatypes = [ObjectDataType]

    class ModelSchema(ma.Schema):
        marshmallow = ma.fields.Nested(ObjectMarshmallowSchema)

    def marshmallow_register_class_names(
        self, *, datatype, classes, marshmallow_module, **kwargs
    ):
        self._register_class_name(
            datatype, datatype.section_marshmallow.config, classes, marshmallow_module
        )

    def marshmallow_build_class_name_existing(
        self, *, datatype, classes, marshmallow_module, **kwargs
    ):
        if datatype.section_marshmallow.config.get("class"):
            self._build_class_name(
                datatype,
                datatype.section_marshmallow.config,
                datatype.definition.setdefault("marshmallow", {}),
                classes,
                marshmallow_module,
                datatype.section_marshmallow.fingerprint,
                "Schema",
            )

    def marshmallow_build_class_name_new(
        self, *, datatype, classes, marshmallow_module, **kwargs
    ):
        if not datatype.section_marshmallow.config.get("class"):
            self._build_class_name(
                datatype,
                datatype.section_marshmallow.config,
                datatype.definition.setdefault("marshmallow", {}),
                classes,
                marshmallow_module,
                datatype.section_marshmallow.fingerprint,
                "Schema",
            )

    def marshmallow_build_class(self, *, datatype, classes, **kwargs):
        self._build_class(
            datatype,
            datatype.section_marshmallow.config,
            datatype.section_marshmallow.children,
            "marshmallow_field",
            classes,
        )

    def marshmallow_field(
        self, datatype: DataType, *, fields: List[MarshmallowField], **kwargs
    ):
        section = datatype.section_marshmallow
        f = []
        super().marshmallow_field(datatype, fields=f)
        fld: MarshmallowField = f[0]
        fld.reference = MarshmallowReference(reference=section.config["class"])
        fields.append(fld)

    def _marshmallow_field_arguments(self, datatype, section, marshmallow, field_name):
        return [
            "__reference__",
            *super()._marshmallow_field_arguments(
                datatype, section, marshmallow, field_name
            ),
        ]



from pathlib import Path

from oarepo_model_builder.builders.python import PythonBuilder
from oarepo_model_builder.datatypes.datatypes import MergedAttrDict
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.utils.dict import dict_get
from oarepo_model_builder.utils.python_name import module_to_path


class InvenioBaseClassPythonBuilder(PythonBuilder):
    section: str
    template: str
    parent_modules = True
    skip_if_not_generating = True

    @property
    def generate(self):
        if hasattr(self, "section") and dict_get(
            self.current_model.definition, [self.section, "skip"], False
        ):
            return False

        if (
            self.skip_if_not_generating
            and hasattr(self, "section")
            and not dict_get(
                self.current_model.definition, [self.section, "generate"], False
            )
        ):
            return False
        return True

    def finish(self, **extra_kwargs):
        super().finish()
        if not self.generate:
            return
        module = self._get_output_module()
        python_path = Path(module_to_path(module) + ".py")

        section = getattr(
            self.current_model,
            f"section_mb_{self.TYPE.replace('-', '_')}",
        )
        merged = MergedAttrDict(section.config, self.current_model.definition)
        self.process_template(
            python_path,
            self.template,
            current_module=module,
            vars=merged,
            **extra_kwargs,
        )

    def _get_output_module(self):
        module = self.current_model.definition[self.section]["module"]
        return module

    def process_template(self, python_path: Path, template, **extra_kwargs):
        if self.parent_modules:
            self.create_parent_modules(python_path)
        output: PythonOutput = self.builder.get_output("python", python_path)
        context = dict(
            settings=self.settings,
            current_model=self.current_model,
            schema=self.current_model.schema.schema,
            **extra_kwargs,
        )
        output.merge(template, context)



from marshmallow import ValidationError

from oarepo_model_builder.validation.model_validation import model_validator


class InvalidModelException(Exception):
    pass


def flatten_errors(err_data, path):
    if isinstance(err_data, (list, tuple)):
        for err in err_data:
            yield from flatten_errors(err, path)
    elif isinstance(err_data, dict):
        for k, v in err_data.items():
            subpath = f"{path}.{k}" if path else k
            yield from flatten_errors(v, subpath)
    else:
        yield str(err_data), path


def validate_model(model):
    try:
        model.schema = model_validator.validate(model.schema)
        return True
    except ValidationError as e:
        msg = []
        for err, path in flatten_errors(e.messages_dict, ""):
            if path.endswith("._schema") and err == "Unknown field.":
                continue
            msg.append(f"{path}: {err}")
        msg = "\n    ".join(msg)
        raise InvalidModelException(f"Invalid model: \n    {msg}")



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import convert_config_to_qualified_name
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .utils import set_default


class BlueprintSchema(ma.Schema):
    generate = ma.fields.Bool(metadata={"doc": "Generate blueprint, defaults to true"})
    alias = ma.fields.Str(
        metadata={
            "doc": "Alias under which the blueprint will be registered in the setup.cfg"
        }
    )
    extra_code = ma.fields.Str(metadata={"doc": "Extra code to be pasted to blueprint"})
    module = ma.fields.Str(metadata={"doc": "Module that will contain the blueprint"})
    function = ma.fields.Str(metadata={"doc": "Fully qualified blueprint function"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema),
        required=False,
        metadata={"doc": "Python imports"},
    )  # imports must be here as well as it is used on model's root (without field)

    class Meta:
        unknown = ma.RAISE


class BlueprintsModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        api_blueprint = ma.fields.Nested(
            BlueprintSchema,
            attribute="api-blueprint",
            data_key="api-blueprint",
            metadata={"doc": "API blueprint details"},
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        alias = datatype.definition["module"]["alias"]
        module = datatype.definition["module"]["qualified"]
        profile = context["profile_module"]

        api = set_default(datatype, "api-blueprint", {})
        api.setdefault("generate", True)
        api.setdefault("alias", alias)
        api.setdefault("extra_code", "")
        api_module = api.setdefault(
            "module",
            f"{module}.views.{profile}.api",
        )
        api.setdefault(
            "function",
            f"{api_module}.create_api_blueprint",
        )
        api.setdefault("imports", [])
        convert_config_to_qualified_name(api, name_field="function")

        app = set_default(datatype, "app-blueprint", {})
        app.setdefault("generate", True)
        app.setdefault("alias", alias)
        app.setdefault("extra_code", "")
        ui_module = app.setdefault(
            "module",
            f"{module}.views.{profile}.app",
        )
        app.setdefault(
            "function",
            f"{ui_module}.create_app_blueprint",
        )
        app.setdefault("imports", [])
        convert_config_to_qualified_name(app, name_field="function")



import marshmallow as ma

from oarepo_model_builder.datatypes import ArrayDataType, DataTypeComponent
from oarepo_model_builder.validation.utils import StrictSchema


class SampleSchema(StrictSchema):
    skip = ma.fields.Boolean(
        metadata={"doc": "Set true to skip generating sample for the field"}
    )
    faker = ma.fields.String(
        metadata={"doc": "The faker to use for generating the sample"}
    )
    params = ma.fields.Raw(metadata={"doc": "Params for the faker"})


class RegularSampleComponent(DataTypeComponent):
    class ModelSchema(ma.Schema):
        sample = ma.fields.Nested(
            SampleSchema,
            required=False,
        )


class ArraySampleSchema(StrictSchema):
    count = ma.fields.Int()


class ArraySampleComponent(RegularSampleComponent):
    eligible_datatypes = [ArrayDataType]

    class ModelSchema(StrictSchema):
        sample = ma.fields.Nested(
            ArraySampleSchema,
            required=False,
        )



import dataclasses
from typing import List

import marshmallow as ma

from ..datatypes import Import, Section, datatypes
from .containers import ObjectDataType


@dataclasses.dataclass
class Link:
    name: str
    link_class: str = "RecordLink"
    link_args: List[str] = dataclasses.field(default_factory=list)
    imports: List[Import] = dataclasses.field(default_factory=list)

    @property
    def link_args_str(self):
        return ", ".join(x for x in self.link_args)


class ModelDataType(ObjectDataType):
    model_type = "model"
    json_schema = {"type": "object"}

    mapping = {}

    class ModelSchema(ObjectDataType.ModelSchema):
        type = ma.fields.Str(
            load_default="model",
            required=False,
            validate=ma.validate.Equal("model"),
        )

    def _process_mapping(self, section: Section, **kwargs):
        section.config = (
            {}
        )  # remove default stuff as mapping on model means not "properties"-level mapping, but root-level mapping
        if self.definition.get("searchable") is False:
            section.config["enabled"] = False
        else:
            section.config.setdefault("enabled", True)

    @property
    def section_global_mapping(self):
        return self.default_section_mapping

    @property
    def links(self):
        return {
            "links_item": [
                Link(
                    name="self",
                    link_class="RecordLink",
                    link_args=['"{self.url_prefix}{id}"'],
                    imports=[Import("invenio_records_resources.services.RecordLink")],
                ),
            ],
            "links_search": [
                Link(
                    name=None,
                    link_class="pagination_links",
                    link_args=['"{self.url_prefix}{?args*}"'],
                    imports=[
                        Import("invenio_records_resources.services.pagination_links")
                    ],
                ),
            ],
        }

    def prepare(self, context):
        datatypes.call_components(
            datatype=self, method="before_model_prepare", context=context
        )
        super().prepare(context)
        datatypes.call_components(
            datatype=self, method="after_model_prepare", context=context
        )



from typing import List

from oarepo_model_builder.datatypes.containers.array import ArrayDataType
from oarepo_model_builder.datatypes.datatypes import DataType, Section, datatypes

from .field import RegularMarshmallowComponent, RegularMarshmallowComponentMixin
from .graph import MarshmallowField


class ArrayMarshmallowComponentMixin(RegularMarshmallowComponentMixin):
    def _create_marshmallow_field(
        self, datatype, section, marshmallow, fields, field_accessor=None, **kwargs
    ):
        f = []
        datatypes.call_components(section.item, field_accessor, fields=f)
        item_field: MarshmallowField = f[0]
        f = []
        super()._create_marshmallow_field(
            datatype=datatype,
            section=section,
            marshmallow=marshmallow,
            fields=f,
            item_field=item_field,
        )
        fld: MarshmallowField = f[0]
        fld.imports.extend(item_field.imports)
        fld.reference = item_field.reference
        fields.append(fld)

    def _marshmallow_field_arguments(
        self,
        datatype,
        section: Section,
        marshmallow,
        field_name,
        *,
        item_field: MarshmallowField = None,
    ):
        args = [item_field.full_definition]
        args.extend(
            super()._marshmallow_field_arguments(
                datatype, section, marshmallow, field_name
            )
        )
        return args


class ArrayMarshmallowComponent(
    ArrayMarshmallowComponentMixin, RegularMarshmallowComponent
):
    eligible_datatypes = [ArrayDataType]

    def marshmallow_field(
        self, datatype: DataType, *, fields: List[MarshmallowField], **kwargs
    ):
        section = datatype.section_marshmallow
        self._create_marshmallow_field(
            datatype,
            section,
            section.config,
            fields,
            field_accessor="marshmallow_field",
            **kwargs,
        )



from ..datatypes import ModelDataType, Section
from ..utils.deepmerge import deepmerge
from ..utils.dict import dict_get
from .json_base import JSONBaseBuilder


def deep_searchable_enabled(dt):
    mapping = dt.section_mapping
    facets = dt.section_facets
    if (
        mapping.config.get("enabled", None) is False
        or facets.config.get("searchable") is False
    ):
        return False
    if mapping.item:
        return deep_searchable_enabled(mapping.item)
    if not mapping.children:
        return True
    for c in mapping.children.values():
        if deep_searchable_enabled(c):
            return True
    return False


class MappingBuilder(JSONBaseBuilder):
    TYPE = "mapping"
    output_file_type = "mapping"
    output_file_name = ["mapping-settings", "file"]
    skip = ["mapping-settings", "skip"]
    parent_module_root_name = "mappings"
    create_parent_packages = True

    def build_node(self, node):
        skip = dict_get(
            self.current_model.definition, ["mapping-settings", "skip"], False
        )
        if skip:
            return
        generated = self.generate_model(node)
        generated.pop("enabled", None)
        self.output.merge(generated)

    def generate_model(self, node):
        generated = self.generate(node)
        generated.pop("enabled", None)
        generated.pop("type", None)

        return {**node.section_global_mapping.config, "mappings": generated}

    def generate(self, node, parent_enabled=True):
        mapping: Section = node.section_mapping
        facets: Section = node.section_facets
        ret = {**mapping.config}
        enabled_property = (
            "enabled" if mapping.config.get("type") in ("object", "nested") else "index"
        )

        searchable = facets.config.get("searchable")
        if searchable is not None:
            ret.setdefault(enabled_property, searchable)

        this_node_enabled = True
        if not isinstance(node, ModelDataType):
            if not deep_searchable_enabled(node):
                ret.setdefault(enabled_property, False)
                return ret

            if not parent_enabled:
                ret.setdefault(enabled_property, False)

            if ret.get(enabled_property) is False:
                return ret
        else:
            this_node_enabled = deep_searchable_enabled(node)

        if mapping.children:
            properties = ret.setdefault("properties", {})
            for k, v in mapping.children.items():
                v = self.generate(v, this_node_enabled)
                if k in properties:
                    deepmerge(properties[k], v)
                else:
                    properties[k] = v
        if mapping.item:
            v = self.generate(mapping.item, this_node_enabled)
            ret.pop("type", None)
            deepmerge(ret, v)
        if ret.get(enabled_property) is True:  # keep only enabled: False there
            ret.pop(enabled_property)
        return ret



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordResourceConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_resource_config"
    section = "resource-config"
    template = "resource-config"



import os
from pathlib import Path

import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.camelcase import camel_case, snake_case
from oarepo_model_builder.utils.python_name import (
    convert_name_to_python,
    split_base_name,
)

from .utils import set_default


class ModuleSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    alias = ma.fields.String(metadata={"doc": "Alias to use for setup.cfg etc."})
    qualified = ma.fields.String(
        metadata={"doc": "Module (fully qualified) where the model is generated"}
    )
    path = ma.fields.String(
        metadata={"doc": "Path on the filesystem with the top-level module"}
    )

    base = ma.fields.String(
        metadata={"doc": "Base name of the module (if the module has dot)"}
    )
    base_upper = ma.fields.String(
        attribute="base-upper",
        data_key="base-upper",
        metadata={"doc": "Uppercase of the base name"},
    )

    kebab_module = ma.fields.String(
        attribute="kebab-module",
        data_key="kebab-module",
        metadata={"doc": "Kebab case of the module"},
    )

    prefix = ma.fields.String(
        metadata={"doc": "Prefix that will be applied to class names"}
    )
    prefix_upper = ma.fields.String(metadata={"doc": "Uppercase variant of the prefix"})
    prefix_snake = ma.fields.String(metadata={"doc": "Snake variant of the prefix"})

    suffix = ma.fields.String(
        metadata={"doc": "Suffix that will be applied to various names"}
    )
    suffix_upper = ma.fields.String(metadata={"doc": "Uppercase variant of the suffix"})
    suffix_snake = ma.fields.String(metadata={"doc": "Snake variant of the suffix"})


class DefaultsModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]

    class ModelSchema(ma.Schema):
        model_name = ma.fields.String(
            attribute="model-name",
            data_key="model-name",
            metadata={
                "doc": "Name of the model, will be used as module name if no module is specified"
            },
        )
        module = ma.fields.Nested(
            ModuleSchema, metadata={"doc": "Model module details"}
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        def get_model_name(_):
            model_name = datatype.definition.get("module", {}).get("qualified")
            if model_name:
                return model_name.rsplit(".", maxsplit=1)[-1].title()
            return os.path.basename(context.get("output-directory", os.getcwd()))

        model_name = set_default(datatype, "model-name", get_model_name)
        model_name = convert_name_to_python(model_name).lower()
        module_container = set_default(datatype, "module", {})

        module_container.setdefault("alias", model_name)
        module = module_container.setdefault("qualified", model_name)

        module_path = module.split(".")
        module_container.setdefault(
            "path", str(Path(module_path[0]).joinpath(*module_path[1:]))
        )

        module_base = module_container.setdefault(
            "base",
            split_base_name(module),
        )
        module_container.setdefault(
            "base-upper",
            module_base.upper(),
        )
        module_container.setdefault(
            "kebab-module",
            module.replace(".", "-").replace("_", "-"),
        )

        prefix = module_container.setdefault(
            "prefix",
            camel_case(module_base),
        )
        module_container.setdefault(
            "prefix-upper",
            prefix.upper(),
        )
        module_container.setdefault(
            "prefix-snake",
            snake_case(prefix),
        )

        suffix = module_container.setdefault(
            "suffix",
            snake_case(prefix),
        )
        module_container.setdefault(
            "suffix-upper",
            suffix.upper(),
        )
        module_container.setdefault(
            "suffix-snake",
            snake_case(suffix),
        )



from .array import ArrayMarshmallowComponent
from .field import (
    MarshmallowField,
    PropertyMarshmallowSchema,
    RegularMarshmallowComponent,
)
from .object import (
    ExtraField,
    ObjectMarshmallowComponent,
    ObjectMarshmallowExtraSchema,
    ObjectMarshmallowMixin,
    ObjectMarshmallowSchema,
)
from .ui_array import UIArrayMarshmallowComponent
from .ui_field import UIMarshmallowComponent
from .ui_object import UIObjectMarshmallowComponent

__all__ = [
    "RegularMarshmallowComponent",
    "MarshmallowField",
    "PropertyMarshmallowSchema",
    "ObjectMarshmallowComponent",
    "ExtraField",
    "ObjectMarshmallowExtraSchema",
    "ObjectMarshmallowMixin",
    "ObjectMarshmallowSchema",
    "ArrayMarshmallowComponent",
    "UIMarshmallowComponent",
    "UIObjectMarshmallowComponent",
    "UIArrayMarshmallowComponent",
]
